@@ -291,11 +291,19 @@ def mk_subtypes(results):
         strain = row["strain_name"]["value"]
 
         if strain not in entries:
-            entry = dict(isolates=set(), ha_subtypes=[], na_subtypes=[], subtypes=[], serotypes=[])
+            entry = dict(
+                isolates=set(),
+                ha_subtypes=[],
+                na_subtypes=[],
+                subtypes=[],
+                serotypes=[],
+            )
         else:
             entry = entries[strain]
 
-        isolates = entry["isolates"].update([i for i in row["isolates"]["value"].split("+") if i])
+        isolates = entry["isolates"].update(
+            [i for i in row["isolates"]["value"].split("+") if i]
+        )
 
         append_add(entry, "genbank_subtypes", default_access(row, "genbank_subtypes"))
         append_add(entry, "gisaid_subtypes", default_access(row, "gisaid_subtypes"))
@@ -310,7 +318,8 @@ def mk_subtypes(results):
 
         entries[strain] = entry
 
-    print("strain_name\tsubtype")
+    strain_entries = []
+    isolate_entries = []
     for strain, entry in entries.items():
         subtype = _get_subtype(
             strain,
@@ -320,10 +329,11 @@ def mk_subtypes(results):
             genbank_subtypes=entry["genbank_subtypes"],
         )
         if subtype:
-          print("\t".join([strain, subtype]))
-          for isolate in entry["isolates"]:
-            print("\t".join([isolate, subtype]))
+            strain_entries.append((strain, subtype))
+            for isolate in entry["isolates"]:
+                isolate_entries.append((isolate, subtype))
 
+    return (strain_entries, isolate_entries)
 
 
 MASTERLIST_HEADER = [
