@@ -1,5 +1,3 @@
-#!/usr/bin/env python3
-
 import click
 import collections
 import os
@@ -7,6 +5,8 @@ import signal
 import sys
 import textwrap
 from octofludb.util import log
+import octofludb.classes as classes
+import octofludb.script as script
 
 
 def open_graph():
@@ -16,7 +16,7 @@ def open_graph():
     return Graph(namespace_manager=manager)
 
 
-def with_graph(f, filename=None, outfile=sys.stdout, *args, **kwargs):
+def with_graph(f, *args, filename=None, outfile=sys.stdout, **kwargs):
     g = open_graph()
 
     if filename is not None:
@@ -120,6 +120,7 @@ def init_cmd(url, repo):
     """
     import pgraphdb as db
     import requests
+    import shutil
 
     config_file = os.path.join(
         os.path.dirname(__file__), "data", "octofludb-config.ttl"
@@ -130,6 +131,183 @@ def init_cmd(url, repo):
         print(f"Could not connect to a GraphDB database at {url}", file=sys.stderr)
         sys.exit(1)
 
+    octofludb_home = os.path.join(os.path.expanduser("~"), ".octofludb")
+
+    if not os.path.exists(octofludb_home):
+        try:
+            print(
+                f" - Creating local configuration folder at '{octofludb_home}'",
+                file=sys.stderr,
+            )
+            os.mkdir(octofludb_home)
+        except FileExistsError:
+            print(
+                f" Failed to create {octofludb_home}",
+                file=sys.stderr,
+            )
+            sys.exit(1)
+
+    script.initialize_config_file()
+
+
+@click.command(
+    name="pull",
+)
+@click.option(
+    "--nmonths",
+    help="Update Genbank files for the last N months",
+    default=3,
+    type=click.IntRange(min=1, max=9999),
+)
+@url_opt
+@repo_name_opt
+def pull_cmd(nmonths, url, repo):
+    """
+    Update data. Pull from genbank, process any new data in the data folder,
+    assign clades to swine data, assign subtypes to all data, and extract
+    motifs
+    """
+    import octofludb.recipes as recipe
+    import pgraphdb as db
+
+    cwd = os.getcwd()
+
+    script.gotoBuildHome()
+
+    config = script.load_config_file()
+
+    # upload ontological schema
+    schema_file = script.get_data_file("schema.ttl")  #
+    upload([schema_file], url=url, repo=repo)
+
+    # upload geological relationships
+    geog_file = upload([script.get_data_file("geography.ttl")], url=url, repo=repo)[0]
+
+    # update genbank (take a parameter telling how far back to go)
+    # this command fills the current directory with .gb* files
+    gb_turtles = prep_update_gb(minyear=1900, maxyear=2121, nmonths=nmonths)
+    upload(gb_turtles, url=url, repo=repo)
+
+    epiflu_metafiles = script.epiflu_meta_files(config)
+    skipped_meta = 0
+    if epiflu_metafiles:
+        for epiflu_metafile in epiflu_metafiles:
+            outfile = os.path.basename(epiflu_metafile) + ".ttl"
+            if os.path.exists(outfile) and os.path.getsize(outfile) > 0:
+                skipped_meta += 1
+            else:
+                with open(outfile, "w") as f:
+                    with_graph(recipe.mk_gis, epiflu_metafile, outfile=f)
+                upload([outfile], url=url, repo=repo)
+    else:
+        log("No epiflu metafiles found")
+    if skipped_meta > 0:
+        log(
+            f"Skipped {str(skipped_meta)} epiflu meta files where existing non-empty turtle files were found in the build directory"
+        )
+
+    epiflu_fastafiles = script.epiflu_fasta_files(config)
+    skipped_fasta = 0
+    if epiflu_fastafiles:
+        for infile in epiflu_fastafiles:
+            outfile = os.path.basename(infile) + ".ttl"
+            if os.path.exists(outfile) and os.path.getsize(outfile) > 0:
+                skipped_fasta += 1
+            else:
+                with open(outfile, "w") as f:
+                    prep_fasta(filename=infile, outfile=f)
+                    upload([outfile], url=url, repo=repo)
+    else:
+        log("No epiflu fasta found")
+
+    if skipped_fasta > 0:
+        log(
+            f"Skipped {str(skipped_fasta)} epiflu fasta files where existing non-empty turtle files were found in the build directory"
+        )
+
+    # octoflu classifications of unclassified swine
+    # * retrieve unclassified strains
+    unclassified_fasta = "unclassified-swine.fna"
+    unclassified_classes = "unclassified-swine.txt"
+    unclassified_turtle = "unclassified-swine.ttl"
+
+    sparql_file = script.get_data_file("fetch-unclassified-swine.rq")
+    with open(unclassified_fasta, "w") as fastaout:
+        fmt_query_cmd(
+            sparql_filename=sparql_file,
+            header=False,
+            fasta=True,
+            url=url,
+            repo=repo,
+            outfile=fastaout,
+        )
+
+    with open(unclassified_classes, "w") as classout:
+        # feed them into runOctoFLU
+        classify_and_write(unclassified_fasta, outfile=classout)
+
+    with open(unclassified_turtle, "w") as turtleout:
+        # print the results
+        prep_table(unclassified_classes, outfile=turtleout)
+
+    upload([unclassified_turtle], url=url, repo=repo)
+
+    # infer subtypes
+    subtypes_table = "subtypes.txt"
+    with open(subtypes_table, "w") as subtypesout:
+        make_subtypes(url=url, repo=repo, outfile=subtypesout)
+
+    genbank_subtypes = "subtypes-genbank.txt"
+    epiflu_subtypes = "subtypes-epiflu.txt"
+    with open(subtypes_table, "r") as subtypesin:
+        with open(genbank_subtypes, "w") as gh:
+            with open(epiflu_subtypes, "w") as eh:
+                for (i, row) in enumerate(subtypesin.readlines()):
+                    row = row.strip()
+                    if i == 0:
+                        print("strain_name\tsubtype", file=gh)
+                        print("isolate_id\tsubtype", file=eh)
+                    else:
+                        if "EPI_ISL" in row:
+                            print(row, file=eh)
+                        else:
+                            print(row, file=gh)
+
+    gturtles = "subtypes-genbank.ttl"
+    eturtles = "subtypes-epiflu.ttl"
+    with open(gturtles, "w") as gturtleout:
+        prep_table(genbank_subtypes, outfile=gturtleout)
+    with open(eturtles, "w") as eturtleout:
+        prep_table(epiflu_subtypes, outfile=eturtleout)
+    upload([gturtles, eturtles], url=url, repo=repo)
+
+    # infer constellations
+    constellation_table = "constellations.txt"
+    constellation_turtles = "constellations.ttl"
+
+    delete_constellations = script.get_data_file("delete-constellations.rq")
+    db.update(sparql_file=delete_constellations, url=url, repo_name=repo)
+
+    with open(constellation_table, "w") as constout:
+        make_const(url=url, repo=repo, outfile=constout)
+
+    with open(constellation_turtles, "w") as turtleout:
+        prep_table(constellation_table, outfile=turtleout)
+
+    upload([constellation_turtles], url=url, repo=repo)
+
+    # load all tags
+    for (tag, basename) in config["tags"].items():
+        for filename in script.tag_files(config, tag):
+            outfile = filename + ".ttl"
+            with open(outfile, "w") as f:
+                prep_tag(tag, filename, outfile=f)
+            upload([outfile], url=url, repo=repo)
+
+    os.chdir(cwd)
+
+    return None
+
 
 @click.command(
     name="clean",
@@ -143,16 +321,17 @@ def clean_cmd(url, repo):
     raise NotImplemented
 
 
-def fmt_query_cmd(sparql_filename, header, fasta, url, repo):
+def fmt_query_cmd(sparql_filename, header, fasta, url, repo, outfile=sys.stdout):
     import octofludb.formatting as formatting
     import pgraphdb as db
 
     results = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo)
     if fasta:
-        formatting.write_as_fasta(results)
+        formatting.write_as_fasta(results, outfile=outfile)
     else:
-        formatting.write_as_table(results, header=header)
-    sys.exit(0)
+        formatting.write_as_table(results, header=header, outfile=outfile)
+
+    return outfile
 
 
 @click.command(
@@ -170,6 +349,43 @@ def query_cmd(*args, **kwargs):
     fmt_query_cmd(*args, **kwargs)
 
 
+@click.command(name="classify")
+@filename_arg
+@click.option("--reference", help="An octoFLU reference fasta file", default=None)
+def classify_cmd(filename, reference=None):
+    """
+    Classify the sequences in a fasta file using octoFLU
+
+    The reference file used will be selected as follows:
+
+      1. If --reference=REFERENCE is given, use REFERENCE. If the file does not exist, die.
+
+      2. If the term `octoflu_references` is defined in the `config.yaml` file
+         in octoflu home, then use this file. If term is not null and does not
+         exist, die.
+
+      3. If no references are given, use the default reference in the octoFLU repo.
+    """
+    classify_and_write(filename, reference=None, outfile=sys.stdout)
+
+
+def classify_and_write(filename, reference=None, outfile=sys.stdout):
+    rows = classify(filename, reference=reference)
+    if rows:
+        print("strain\tsegment\tus_clade\tglobal_clade", file=outfile)
+        for row in rows:
+            print("\t".join(row), file=outfile)
+    else:
+        die(colors.bad("No result from running octFLU"), file=outfile)
+
+
+def classify(filename, reference=None):
+    if not reference:
+        config = script.load_config_file()
+        reference = script.get_octoflu_reference(config)
+    return script.runOctoFLU(filename, reference)
+
+
 @click.command(
     name="construct",
 )
@@ -185,7 +401,8 @@ def construct_cmd(sparql_filename, url, repo):
     results = db.sparql_construct(sparql_file=sparql_filename, url=url, repo_name=repo)
 
     print(results)
-    sys.exit(0)
+
+    return None
 
 
 @click.command(
@@ -214,29 +431,19 @@ def upload_cmd(turtle_filenames, url, repo):
     """
     Upload one or more turtle files to the database
     """
-    import shutil
+    upload(turtle_filenames, url, repo)
+
+
+def upload(turtle_filenames, url, repo):
     import pgraphdb as db
 
-    for filename in turtle_filenames:
-        dbdir = os.path.join(os.path.expanduser("~"), "graphdb-import")
-        if not os.path.exists(dbdir):
-          os.mkdir(dbdir)
-        new_filename = os.path.join(dbdir, os.path.basename(filename)
-        )
-        if filename == new_filename:
-            continue
-        else:
-            try:
-                shutil.copyfile(filename, new_filename)
-            except AttributeError as e:
-                print(
-                    f"Could not move {filename} to {new_filename}: {str(e)}",
-                    file=sys.stderr,
-                )
-                sys.exit(1)
-    server_files = [os.path.basename(f) for f in turtle_filenames]
-    db.load_data(url=url, repo_name=repo, turtle_files=server_files)
-    sys.exit(0)
+    files = []
+    for filenames in turtle_filenames:
+        for filename in script.expandpath(filenames):
+            log(f"loading file: {filename}")
+            db.load_data(url=url, repo_name=repo, turtle_file=filename)
+            files.append(filename)
+    return files
 
 
 # ===== prep subcommands ====
@@ -247,10 +454,14 @@ def upload_cmd(turtle_filenames, url, repo):
 )
 @click.argument("tag", type=str)
 @filename_arg
-def prep_tag_cmd(tag, filename):
+def prep_tag_cmd(tag, filename, outfile=sys.stdout):
     """
     Associate list of IDs with a tag
     """
+    prep_tag(tag, filename, outfile=outfile)
+
+
+def prep_tag(tag, filename, outfile=sys.stdout):
     from rdflib import Literal
     import datetime as datetime
     from octofludb.nomenclature import make_uri, make_tag_uri, P
@@ -270,7 +481,7 @@ def prep_tag_cmd(tag, filename):
         turtles = g.serialize(format="turtle")
         log("done")
         for l in turtles.splitlines():
-            print(l.decode("utf-8"))
+            print(l.decode("utf-8"), file=outfile)
     g.close()
 
 
@@ -369,20 +580,32 @@ def prep_update_gb_cmd(minyear, maxyear, nmonths):
     """
     Retrieve any missing genbank records. Results are stored in files with the prefix '.gb_###.ttl'
     """
+    return prep_update_gb(minyear, maxyear, nmonths)
+
+
+def prep_update_gb(minyear, maxyear, nmonths):
     from octofludb.entrez import missing_acc_by_date
     import octofludb.colors as colors
 
+    outfiles = []
+
     for date, missing_acc in missing_acc_by_date(
         min_year=minyear, max_year=maxyear, nmonths=nmonths
     ):
         if missing_acc:
-            log(colors.good(f"Updating {date} ..."))
             outfile = ".gb_" + date.replace("/", "-") + ".ttl"
-            with open(outfile, "w") as fh:
-                with_graph(_mk_gbids_cmd, gbids=missing_acc, outfile=fh)
+            if os.path.exists(outfile):
+                log(f"GenBank turtle file for '{str(date)}' already exists, skipping")
+            else:
+                log(colors.good(f"Updating {date} ..."))
+                with open(outfile, "w") as fh:
+                    with_graph(_mk_gbids_cmd, gbids=missing_acc, outfile=fh)
+                outfiles.append(outfile)
         else:
             log(colors.good(f"Up-to-date for {date}"))
 
+    return outfiles
+
 
 @click.command(
     name="blast",
@@ -436,17 +659,32 @@ na_opt = click.option("--na", help="The string that represents a missing value")
 @click.option("--levels", help="levels")
 @na_opt
 @segment_key_opt
-def prep_table_cmd(filename, tag, include, exclude, levels, na, segment_key):
+def prep_table_cmd(*args, **kwargs):
+    """
+    Translate a table to RDF
+    """
+    prep_table(*args, **kwargs)
+
+
+def prep_table(
+    filename,
+    tag=None,
+    include=None,
+    exclude=None,
+    levels=None,
+    na=None,
+    segment_key=None,
+    outfile=sys.stdout,
+):
     """
     Translate a table to RDF
     """
-    from octofludb.classes import Table
     from octofludb.recipes import IrregularSegmentTable
 
     if segment_key:
         constructor = IrregularSegmentTable
     else:
-        constructor = Table
+        constructor = classes.Table
 
     def _mk_table_cmd(g, fh):
         (inc, exc, levelsProc) = process_tablelike(include, exclude, levels)
@@ -460,7 +698,7 @@ def prep_table_cmd(filename, tag, include, exclude, levels, na, segment_key):
             na_str=make_na(na),
         ).connect(g)
 
-    with_graph(_mk_table_cmd, filename)
+    with_graph(_mk_table_cmd, filename=filename, outfile=outfile)
 
 
 @click.command(
@@ -472,14 +710,25 @@ def prep_table_cmd(filename, tag, include, exclude, levels, na, segment_key):
 @include_opt
 @exclude_opt
 @na_opt
-def prep_fasta_cmd(filename, tag, delimiter, include, exclude, na):
+def prep_fasta_cmd(*args, **kwargs):
     """
     Translate a fasta file to RDF.
 
     <filename> Path to a TAB-delimited or excel table
     """
-    import octofludb.classes as classes
 
+    return prep_fasta(*args, **kwargs)
+
+
+def prep_fasta(
+    filename,
+    tag=None,
+    delimiter=None,
+    include=None,
+    exclude=None,
+    na=None,
+    outfile=sys.stdout,
+):
     def _mk_fasta_cmd(g, fh):
         (inc, exc, levels) = process_tablelike(include, exclude, None)
         classes.Ragged(
@@ -492,7 +741,9 @@ def prep_fasta_cmd(filename, tag, delimiter, include, exclude, na):
             na_str=make_na(na),
         ).connect(g)
 
-    with_graph(_mk_fasta_cmd, filename)
+    with_graph(_mk_fasta_cmd, filename, outfile=outfile)
+
+    return outfile
 
 
 @click.command(
@@ -594,12 +845,16 @@ def make_const_cmd(url, repo):
     only a few of these in the US. "X" represents something else exotic (e.g.,
     not US). For mixed strains, the constellation will be recorded as "mixed".
     """
+    make_const(url, repo)
+
+
+def make_const(url, repo, outfile=sys.stdout):
     import octofludb.formatting as formatting
     import pgraphdb as db
 
     sparql_filename = os.path.join(os.path.dirname(__file__), "data", "segments.rq")
     results = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo)
-    formatting.write_constellations(results)
+    formatting.write_constellations(results, outfile=outfile)
 
 
 @click.command(
@@ -607,10 +862,21 @@ def make_const_cmd(url, repo):
 )
 @url_opt
 @repo_name_opt
-def make_subtypes_cmd(url, repo):
+def make_subtypes_cmd(url, repo, outfile=sys.stdout):
     """
     Determine subtypes based on Genbank serotype field, epiflu data, or octoflu HA/NA classifications"
     """
+    make_subtypes(url=url, repo=repo, outfile=outfile)
+
+
+def make_subtypes(url, repo, outfile=sys.stdout):
+    strains, isolates = get_missing_subtypes(url, repo)
+    print("strain_name\tsubtype", file=outfile)
+    for identifier, subtype in strains + isolates:
+        print(f"{identifier}\t{subtype}", file=outfile)
+
+
+def get_missing_subtypes(url, repo):
     import octofludb.recipes as recipe
     import pgraphdb as db
 
@@ -618,7 +884,7 @@ def make_subtypes_cmd(url, repo):
 
     results = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo)
 
-    recipe.mk_subtypes(results)
+    return recipe.mk_subtypes(results)
 
 
 @click.command(
@@ -697,6 +963,8 @@ def fetch_tag_cmd(filename, url, repo):
     upload_cmd([turtle_filename], url, repo)
     g.close()
 
+    return None
+
 
 @click.command(
     name="isolate",
@@ -710,7 +978,7 @@ def fetch_isolate_cmd(url, repo):
     sparql_filename = os.path.join(
         os.path.dirname(__file__), "data", "get-tagged-isolate.rq"
     )
-    fmt_query_cmd(sparql_filename, True, False, url, repo)
+    fmt_query_cmd(sparql_filename, header=True, fasta=False, url=url, repo=repo)
 
 
 @click.command(
@@ -725,7 +993,7 @@ def fetch_strain_cmd(url, repo):
     sparql_filename = os.path.join(
         os.path.dirname(__file__), "data", "get-tagged-strain.rq"
     )
-    fmt_query_cmd(sparql_filename, True, False, url, repo)
+    fmt_query_cmd(sparql_filename, header=True, fasta=False, url=url, repo=repo)
 
 
 @click.command(
@@ -740,7 +1008,7 @@ def fetch_segment_cmd(url, repo):
     sparql_filename = os.path.join(
         os.path.dirname(__file__), "data", "get-tagged-segment.rq"
     )
-    fmt_query_cmd(sparql_filename, True, False, url, repo)
+    fmt_query_cmd(sparql_filename, header=True, fasta=False, url=url, repo=repo)
 
 
 @click.command(
@@ -755,7 +1023,7 @@ def fetch_sequence_cmd(url, repo):
     sparql_filename = os.path.join(
         os.path.dirname(__file__), "data", "get-tagged-sequence.rq"
     )
-    fmt_query_cmd(sparql_filename, False, True, url, repo)
+    fmt_query_cmd(sparql_filename, header=False, fasta=True, url=url, repo=repo)
 
 
 @click.command(
@@ -825,10 +1093,11 @@ def macro_query(filename, macros, *args, **kwargs):
                     line = re.sub(macro, replacement, line)
                 print(line, file=query)
 
-    fmt_query_cmd(tmpfile, *args, **kwargs)
+    fmt_query_cmd(sparql_filename, *args, **kwargs)
 
     os.remove(tempfile)
 
+
 @click.command(
     name="monthly",
 )
@@ -843,10 +1112,10 @@ def report_monthly_cmd(year, month, context, url, repo):
     """
 
     def pad(x):
-      if x < 10 and x >= 0:
-        return ("0" + str(x))
-      else:
-        return str(x)
+        if x < 10 and x >= 0:
+            return "0" + str(x)
+        else:
+            return str(x)
 
     if context:
 
@@ -988,9 +1257,11 @@ def cli_grp():
 
 
 cli_grp.add_command(init_cmd)
+cli_grp.add_command(pull_cmd)
 cli_grp.add_command(clean_cmd)
 cli_grp.add_command(query_cmd)
 cli_grp.add_command(update_cmd)
+cli_grp.add_command(classify_cmd)
 cli_grp.add_command(construct_cmd)
 cli_grp.add_command(upload_cmd)
 cli_grp.add_command(prep_grp)
